# -*- coding: ISO-8859-1 -*-
# Copyright (C) 2003-2005 Juan David Ibez Palomar <jdavid@itaapy.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# Import from the Standard Library
from base64 import decodestring
from copy import copy
import cStringIO
import datetime
import logging
import os
from popen2 import popen3
import tempfile
from time import time
from urllib import quote, unquote

# Import from itools
from itools import get_abspath
from itools import types
from itools.resources import zope2, get_resource
from itools import handlers
from itools.handlers import get_handler
from itools.handlers.KeyValue import KeyValue
from itools.handlers import transactions
from itools.xml import namespaces
from itools.catalog.Catalog import Catalog
from itools.zope import get_context

# Import from ikaaro
from Group import Group
from Handler import Handler
from exceptions import Forbidden, UserError
from Metadata import Metadata, Record
from text import PO
from skins import ui
from User import User
from UserFolder import UserFolder
from utils import comeback
from WebSite import WebSite

# Initialize the local logger for error_page
error_page_logger = logging.getLogger('error_page')
error_page_logger.setLevel(logging.ERROR)
error_page_logger_buffer = cStringIO.StringIO()
error_page_logger_handler = logging.StreamHandler(error_page_logger_buffer)
error_page_logger.addHandler(error_page_logger_handler)



class Settings(KeyValue):

    __keys__ = ['join', 'languages']
    __keys_types__ = {'join': 'bool'}


    def _load_state(self, resource):
        KeyValue._load_state(self, resource)

        state = self.state
        state.languages = state.languages.split()
        state.default_language = state.languages and state.languages[0] or None



class Root(Group, WebSite):

    class_id = 'iKaaro'
    class_title = u'iKaaro'
    class_version = '20050509'

    __fixed_handlers__ = ['users', 'admins', 'reviewers', 'ui']


    ########################################################################
    # Skeleton
    ########################################################################
    _catalog_fields = [('text', 'text', True, False),
                       ('title', 'text', True, True),
                       ('owner', 'keyword', True, True),
                       ('is_group', 'bool', True, False),
                       ('format', 'keyword', True, True),
                       ('workflow_state', 'keyword', True, True),
                       ('abspath', 'keyword', True, True),
                       ('usernames', 'keyword', True, False),
                       # Folder's view
                       ('parent_path', 'keyword', True, True),
                       ('name', 'keyword', True, True),
                       ('title_or_name', 'keyword', True, True),
                       ('mtime_microsecond', 'keyword', False, True),
                       ]


    def get_skeleton(self, username=None, password=None):
        skeleton = {}
        # The catalog must be added first, so everything else is indexed
        # (<index name>, <type>, <indexed>, <stored>)
        catalog = Catalog(fields=self._catalog_fields)
        skeleton['.catalog'] = catalog
        # The archive is used for versioning
        skeleton['.archive'] = handlers.Folder.Folder()
        # Now call the parents get_skeleton
        users = [username]
        skeleton.update(Group.get_skeleton(self, users=users))
        # Reviewers
        skeleton['reviewers'] = Group(users=users)
        metadata = Metadata(handler_class=Group, owner=username,
                            title={'en': u'Reviewers'})
        skeleton['.reviewers.metadata'] = metadata
        # Admins
        skeleton['admins'] = Group(users=users)
        metadata = Metadata(handler_class=Group, owner=username,
                            title={'en': u'Admins'})
        skeleton['.admins.metadata'] = metadata
        # Metadata
        skeleton['.metadata'] = Metadata(handler_class=self.__class__)
        # Users
        skeleton['users'] = UserFolder(users=[(username, password)])
        metadata = Metadata(handler_class=UserFolder, owner=username,
                            title={'en': u'Users'})
        skeleton['.users.metadata'] = metadata
        # Message catalog
        skeleton['en.po'] = PO()
        skeleton['.en.po.metadata'] = Metadata(handler_class=PO)
        # That's all
        return skeleton


    def get_catalog_metadata_fields(self):
        return [field[0] for field in self._catalog_fields if field[3]]


    ########################################################################
    # Publish
    ########################################################################
    def publish(self):
        context = get_context()
        request, response = context.request, context.response
        path = context.path

        # Set the application's root in the context
        context.root = self
        segment, path = path[0], path[1:]
        self.name = segment.name

        # Create the security context
        cname = '__ac'
        cookie = request.cookies.get(cname, None)
        if cookie is not None:
            cookie = unquote(cookie)
            cookie = decodestring(cookie)
            username, password = cookie.split(':', 1)
            # 
            user = self.get_user(username)
            if user.authenticate(password):
                context.user = user

        # Traverse
        self.before_traverse()
        handler = self.get_handler(path)
        context.handler = handler

        # Get the method
        method_name = context.method
        http_method = request.environ.get('REQUEST_METHOD')

        if method_name is None:
            method_name = http_method

        method = handler.get_method(method_name)

        # Check security
        if method is None:
            method = handler.forbidden_form

        # Set the list of needed resources. The method we are going to
        # call may need external resources to be rendered properly, for
        # example it could need an style sheet or a javascript file to
        # be included in the html head (which it can not control). This
        # attribute lets the interface to add those resources.
        context.styles = []
        context.scripts = []

        # Build parameters
        if method.im_func.func_code.co_flags & 8:
            parameters = {}
            for key, value in request.form.items():
                n = key.count('.')
                if n == 0:
                    prefix = None
                    local_name = key
                    language = None
                if n == 1:
                    prefix, local_name = key.split('.')
                    language = None
                elif n == 2:
                    prefix, local_name, language = key.split('.')

                if prefix is not None:
                    namespace = namespaces.get_namespace_by_prefix(prefix)
                    schema = namespace.get_attribute_schema(local_name)
                    type = schema['type']
                    value = type.decode(value)
                    key = '%s:%s' % (prefix, local_name)

                if language is None:
                    parameters[key] = value
                else:
                    parameters.setdefault(key, {})[language] = value
        else:
            parameters = None

        try:
            # Call the method
            if parameters is None:
                response_body = method()
            else:
                response_body = method(**parameters)
        except UserError, exception:
            message = exception.args[0]
            if isinstance(message, unicode):
                message = message.encode('utf8')
            message = quote(message)
            referer = copy(request.referer)
            referer.query = ''
            url = '%s?message:utf8:ustring=%s' % (referer, message)
            response.redirect(url)
        except Forbidden:
            # XXX
            path = ('../' * (len(path))) + ';login_form'
            response.redirect(path)
            response.set_status(403)
        except:
            # Rollback changes to handlers
            transaction = transactions.get_transaction()
            transaction.rollback()
            if self.has_handler('.catalog'):
                catalog = self.get_handler('.catalog')
                catalog.timestamp = datetime.datetime(1900, 1, 1)
            # Abort Zope's transaction
            get_transaction().abort()
            # Log error
            error_log = self.log_error()
            error_log = quote(error_log)
            # Redirect to error page
            # XXX
            path = ('../' * (len(path))) 
            url = path + ';error_page?error_log=%s' % error_log
            response.redirect(url)
        else:
            # Save changes
            transaction = transactions.get_transaction()
            transaction.commit()
            # If there is not content type and the body is not None,
            # wrap it in the skin template
            if http_method == 'GET' and response_body is not None:
                if not response.has_header('content-type'):
                    skin = self.get_skin()
                    response_body = skin.template(response_body)

                # Set the response body
                response.body = response_body

            # For now the response is in the request
            response = context.response


    def log_error(self):
        context = get_context()
        request, user = context.request, context.user
        # Log error
        error = '\n' \
                '[iKaaro Error]\n' \
                'date   : %(date)s\n' \
                'uri    : %(uri)s\n' \
                'referer: %(referer)s\n' \
                'user   : %(user)s\n' \
                '\n'
        user = context.user
        error = error % {'date': str(datetime.datetime.now()),
                         'uri': str(request.uri),
                         'referer': str(request.referer),
                         'user': user and user.name or None}
        # log on the root logger (std_out and ./log/ikaaro.log)
        logging.getLogger().exception(error)

        # log on the local logger
        error_page_logger.exception(error)
        traceback = error_page_logger_buffer.getvalue()

        # empty the error_page_logger_buffer
        error_page_logger_buffer.truncate(0)
        return traceback

    ########################################################################
    # Traverse
    ########################################################################
    def _get_handler_names(self):
        return Group._get_handler_names(self) + ['ui']


    def _get_handler(self, segment, resource):
        name = segment.name

        # XXX To be removed by 0.15
        if name == '.settings':
            return Settings(resource)

        if name == '.catalog':
            return Catalog(resource)
        elif name == '.archive':
            return handlers.Folder.Folder(resource)
        return Group._get_handler(self, segment, resource)


    def _get_virtual_handler(self, segment):
        name = segment.name
        if name == 'ui':
            return ui
        return Group._get_virtual_handler(self, segment)


    ########################################################################
    # Zope registry metatype (replaces bootstrap)
    ########################################################################
    def before_traverse(self):
        context = get_context()
        # Language negotiation
        user = context.user
        if user is not None:
            language = user.get_property('ikaaro:user_language')
            context.request.accept_language.set(language, 2.0)


    @classmethod
    def get_dummy_user(cls, username, password):
        return User(password=password)


    @classmethod
    def manage_addForm(cls, container):
        # Load the template
        path = get_abspath(globals(), 'zmi/add.xml')
        template = get_handler(path)
        # Build the namespace
        namespace = {}
        namespace['title'] = cls.class_title
        namespace['message'] = get_context().request.form.get('message')
        # Process the template
        return template.stl(namespace)


    @classmethod
    def manage_add(cls, container):
        context = get_context()
        request, response = context.request, context.response
        # Pre-process input parameters
        form = request.form
        name = form['name'].strip()
        title = form['title']
        username = form['username'].strip()
        password = form['password']
        password2 = form['password2']
        # Check the input data
        error_message = None
        if not name:
            error_message = 'You must enter the name'
        elif not username:
            error_message = 'You must enter the username'
        elif not password:
            error_message = 'You must enter the password'
        elif not password2:
            error_message = 'You must enter the password twice'
        elif password != password2:
            error_message = 'The passwords differ'

        if error_message is not None:
            referer = request.referer
            response.redirect('%s?message=%s' % (referer.path, error_message))
            return None, None

        # Create new handler
        root_handler = cls(username=username, password=password)

        # Build the user handler, and set it in the request
        user = cls.get_dummy_user(username, password)
        context.user = user

        # Index everything
        catalog = root_handler.get_handler('.catalog')
        for handler, context in root_handler.traverse2():
            abspath = handler.get_abspath()
            if handler.name.startswith('.'):
                context.skip = True
            elif abspath == '/ui':
                context.skip = True
            elif isinstance(handler, Handler):
                print handler.get_abspath()
                catalog.index_document(handler)
        transaction = transactions.get_transaction()
        transaction.commit()

        # Copy the handler tree from memory the Zope 2 database
        container_path = '/'.join(container.getPhysicalPath()) or '/'
        container_resource = zope2.get_resource(container_path)
        container_resource.set_resource(name, root_handler.resource)

        # My little hack
        from Products.ikaaro import __before_publishing_traverse__, \
             __browser_default__
        folder = container._getOb(name)
        folder.__before_publishing_traverse__ = __before_publishing_traverse__
        folder.__browser_default__ = __browser_default__

        # Come back
        response.redirect('manage_main')


    metatypes_registry = {}


    @classmethod
    def register_in_zope(cls, context):
        cls.metatypes_registry[cls.class_id] = cls

        def manage_addForm(container, cls=cls):
            """ """
            return cls.manage_addForm(container)

        def manage_add(container, cls=cls):
            """ """
            cls.manage_add(container)

        context.registerClass(cls, meta_type=cls.class_id,
                              constructors=(manage_addForm, manage_add))


    ########################################################################
    # API
    ########################################################################
    def get_metadata(self):
        return self.get_handler('.metadata')

    metadata = property(get_metadata, None, None, "")


    def get_user(self, name):
        return self.get_handler('users/%s' % name)


    def get_usernames(self):
        return self.get_handler('users').get_usernames()


    def get_groups(self):
        """
        Returns a list with all the subgroups, including the subgroups of
        the subgroups, etc..
        """
        groups = self.search(is_group=True)
        groups = [ x.abspath for x in groups if x.abspath != '/' ]
        return groups


    def get_document_types(self):
        return Group.get_document_types(self) ##+ [WebSite]


    ########################################################################
    # Index & Search
    def index_handler(self, handler):
        catalog = self.get_handler('.catalog')
        n = catalog.index_document(handler)


    def unindex_handler(self, handler):
        catalog = self.get_handler('.catalog')

        abspath = handler.get_abspath()
        for document in catalog.search(abspath=abspath):
            catalog.unindex_document(document.__number__)


    def reindex_handler(self, handler):
        self.unindex_handler(handler)
        self.index_handler(handler)


    ########################################################################
    # Skins and themes (themes are a sub-set of the skins)
    def get_themes(self):
        return ['classic', 'surf']


    def get_skin(self):
        context = get_context()

        # Check the request
        form = context.request.form
        if form.has_key('skin_path'):
            return self.get_handler(form['skin_path'])

        # Check the user preferences
        user = context.user
        if user is None:
            return self.get_handler('ui/surf')

        theme = user.get_property('ikaaro:user_theme')
        return self.get_handler('ui/%s' % theme)


    ########################################################################
    # Settings
    def get_available_languages(self):
        return ['en', 'es', 'fr', 'zh', 'it']


    def get_default_language(self):
        return 'en'

    ########################################################################
    # Back Office
    ########################################################################
    def get_views(self):
        user = get_context().user
        if user is None:
            return ['about', 'login_form', 'join_form']
        return ['browse_thumbnails', 'new_resource_form',
                'edit_metadata_form', 'general_form', 'catalog_form', 'about']


    def get_subviews(self, name):
        views = [['browse_thumbnails', 'browse_list'],
                 ['general_form', 'languages_form'],
                 ['about', 'license']]
        for subviews in views:
            if name in subviews:
                return subviews
        return Group.get_subviews(self, name)


    browse_thumbnails__label__ = u'View'


    ########################################################################
    # About
    about__access__ = True
    about__label__ = u'About'
    about__sublabel__ = u'iKaaro'
    def about(self):
        handler = self.get_handler('/ui/Root_about.xml')
        return handler.stl()


    ########################################################################
    # License
    license__access__ = True
    license__label__ = u'About'
    license__sublabel__ = u'License'
    def license(self):
        handler = self.get_handler('/ui/Root_license.xml')
        return handler.stl()


    ########################################################################
    # Join
    def is_allowed_to_join(self):
        context = get_context()
        website_is_open = context.root.get_property('ikaaro:website_is_open')
        return website_is_open


    join_form__access__ = is_allowed_to_join
    join_form__label__ = u'Join'
    def join_form(self):
        handler = self.get_handler('/ui/Root_join.xml')
        return handler.stl()


    join__access__ = is_allowed_to_join
    def join(self, username, password, password2):
        users = self.get_handler('users')
        error = users.new_user(username, password, password2)

        response = get_context().response
        message = self.gettext('Thanks for joining us, please log in')
        response.redirect(';login_form?username=%s&message:utf8:ustring=%s'
                          % (username, quote(message.encode('utf8'))))


    ########################################################################
    # Catalog
    ########################################################################
    catalog_form__access__ = Handler.is_admin
    catalog_form__label__ = u'Catalog'
    def catalog_form(self):
        handler = self.get_handler('/ui/Root_catalog.xml')
        return handler.stl()


    update_catalog__access__ = Handler.is_admin
    def update_catalog(self):
        # Initialize a new empty catalog
        t0 = time()
        tmp_path = tempfile.mkdtemp()
        tmp_folder = get_handler(tmp_path)
        tmp_folder.set_handler('catalog', Catalog(fields=self._catalog_fields))
        tmp_folder.save_state()
        catalog_resource = tmp_folder.resource.get_resource('catalog')
        catalog = Catalog(catalog_resource)
        # Remove old catalog
        self.del_handler('.catalog')
        t1 = time()
        print 'Updating catalog, initialization:', t1 - t0

        n = 0
        for handler, context in self.traverse2():
            name = handler.name
            abspath = handler.get_abspath()

            if name.startswith('.'):
                context.skip = True
            elif abspath == '/ui':
                context.skip = True
            elif handler.real_handler is not None and not abspath == '/ui':
                context.skip = True
            elif not name.startswith('.'):
                print n, abspath
                catalog.index_document(handler)
                n += 1
                # Avoid too much memory usage but saving changes
                if n % 500 == 0:
                    catalog.save_state()
        catalog.save_state()

        t2 = time()
        print 'Updating catalog, indexing:', t2 - t1
        # Set new catalog
        self.set_handler('.catalog', catalog)
        t3 = time()
        print 'Updating catalog, saving:', t3 - t2
        path = tmp_path.split('/')
        path, name = '/'.join(path[:-1]), path[-1]
        get_resource(path).del_resource(name)
        t4 = time()
        print 'Updating catalog, removing temporary files:', t4 - t3
        print 'Updating catalog, total:', t4 - t0

        message = self.gettext('%s handlers have been re-indexed.') % n
        comeback('catalog_form', message)


##    export_catalog__access__ = True
##    def export_catalog(self):
##        from itools.handlers import get_handler
##        tmp = get_handler('/tmp')
##        catalog = self.get_handler('.catalog')
##        tmp.set_handler('catalog', catalog)
##        tmp.save_state()

##        return 'ok'


    #######################################################################
    # Update
    #######################################################################
    def update_20050509(self):
        # Move settings to metadata
        settings = self.get_handler('.settings')
        if settings.state.join is True:
            self.set_property('ikaaro:website_is_open', True)
        self.set_property('ikaaro:website_languages',
                          tuple(settings.state.languages))
        self.metadata.save_state()

        self.resource.del_resource('.settings')


    #######################################################################
    # Import
    #######################################################################
    ximport__access__ = Handler.is_admin
    def ximport(self, path):
        from itools.handlers import get_handler
        for resource_name in self.resource.get_resource_names():
            self.resource.del_resource(resource_name)
        source = get_resource(path)
        for resource_name in source.get_resource_names():
            resource = source.get_resource(resource_name)
            self.resource.set_resource(resource_name, resource)
        print 'importing, done'
